/*  control.c -- user interaction and control panel window  */

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include "topo.h"

static void reshapeControlPanel(int w, int h);
static void emitControlPanel(void);
static void controlPanelMouseFunc(int button0, int state0, int x0, int y0);
static void mmove(double step, double turn);
static void setEye(int up);
static void emitString(int x, int y, char *s);




/* control panel size */
#define CONTROL_WIDTH	160
#define CONTROL_HEIGHT	256

/* control panel layout */
#define XLAB		10	/* x position for label */
#define XVAL		100	/* x position for value */
#define LSP		18	/* line spacing (leading) */
#define DSC		4	/* descent below baseline */
#define FONT		GLUT_BITMAP_HELVETICA_12

/* redisplay flags */
#define RC		 1	/* redisplay control window */
#define RO		 2	/* redisplay overhead view */
#define RP		 4	/* redisplay perspective view */



/*  control panel fields  */
#define CONTROLLABELLENGTH 20

struct {
  unsigned char cl;	/* keyboard character on left mouse click */
  unsigned char cr;	/* keyboard character on right mouse click */
  int variable;	/* is this field variable? */
  int refresh;	/* which windows to refresh */
  int x, y;		/* location in window */
  char label[CONTROLLABELLENGTH+1];	/* current label value */
} controls[] = {
  /*keyL keyR act  refresh  xposn  yposn    string*/
  { 'c', 'C',  0, RC+RO+RP, XLAB, 1 * LSP, "[C]olor by" },
  { 'c', 'C',  1, RC+RO+RP, XVAL, 1 * LSP, "" },
  { '>', '<',  0, RC+RO+RP, XLAB, 2 * LSP, "Error[>0<]" },
  { '>', '<',  1, RC+RO+RP, XVAL, 2 * LSP, "" },
  { '0', '0',  1, RC+RO+RP, XVAL, 2 * LSP, "" },
  { '+', '-',  0, RC+RO+RP, XLAB, 3 * LSP, "Distance[+-]" },
  { '+', '-',  1, RC+RO+RP, XVAL, 3 * LSP, "" },
  { 'e', 'E',  0, RC+RO+RP, XLAB, 5 * LSP, "[E]ye" },
  { 'e', 'E',  1, RC+RO+RP, XVAL, 5 * LSP, "" },
  { '2', '@',  0, RC+RO   , XLAB, 6 * LSP, "[2]d Mesh" },
  { '2', '@',  1, RC+RO   , XVAL, 6 * LSP, "" },
  { '3', '#',  0, RC   +RP, XLAB, 7 * LSP, "[3]d Mesh" },
  { '3', '#',  1, RC   +RP, XVAL, 7 * LSP, "" },
  { 'v', 'V',  0, RC+RO+RP, XLAB, 8 * LSP, "[V]ert.Exag." },
  { 'v', 'V',  1, RC+RO+RP, XVAL, 8 * LSP, "" },
  { 'b', 'B',  0, RC+RO+RP, XLAB, 9 * LSP, "[B]ackface Cull" },
  { 'b', 'B',  1, RC+RO+RP, XVAL, 9 * LSP, "" },
  { 'f', 'F',  0, RC+RO+RP, XLAB,10 * LSP, "[F]ov Cull" },
  { 'f', 'F',  1, RC+RO+RP, XVAL,10 * LSP, "" },
  { 'p', 'P',  0,        0, XLAB,12 * LSP, "[P]rint" },
  { 'q', 'Q',  0,        0, XLAB,13 * LSP, "[Q]uit" }
};

#define NUMBER_OF_CONTROLS (sizeof(controls) / sizeof(controls[0]))



/*  openCtlWin() -- open control panel window  */

void openCtlWin(void)
{
  char wname[200];

  sprintf(wname, "%s", gemname);
  glutInitDisplayMode(GLUT_RGB | GLUT_SINGLE);
  glutInitWindowSize(CONTROL_WIDTH, CONTROL_HEIGHT);
  ctlWin = glutCreateWindow(wname);
  glutSetWindowTitle(wname);
  glClearColor((GLclampf)1,(GLclampf)1,(GLclampf)1,(GLclampf)1);
  glColor3d((GLdouble)0,(GLdouble)0,(GLdouble)0);

  glutMouseFunc(controlPanelMouseFunc);
  glutDisplayFunc(emitControlPanel);
  glutReshapeFunc(reshapeControlPanel);
  glutKeyboardFunc(keyfunc);
  glutSpecialFunc(specfunc);
}



/*  reshapeControlPanel(w, h) -- set rendering parameters for control panel  */

static void reshapeControlPanel(int w, int h)
{
  glViewport((GLint)0,(GLint)0,(GLint)w,(GLint)h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D((GLdouble)0,(GLdouble)w,(GLdouble)0,(GLdouble)h);
  glScaled((GLdouble)1,(GLdouble)-1,(GLdouble)1);
  glTranslated((GLdouble)0,(GLdouble)-h,(GLdouble)0);
  glMatrixMode(GL_MODELVIEW);
}



/*  emitControlPanel() -- display contents of control panel  */

static void emitControlPanel(void)
{
  int i;
  char *s;

  glClear(GL_COLOR_BUFFER_BIT);

  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (controls[i].x == XVAL) {
      s = controls[i].label;

      switch(controls[i].cl) {

      case '+':
	sprintf(s, "%4.1fm", distThresh);  break;
	break;

      case '>':
	sprintf(s, "%4.2fm/km", errThresh*1000.);  break;
	break;

      case 'b':
	strcpy(s, cullBackface ? "On" : "Off");  break;

      case 'c':
	switch(colorLayer) {
	case T_CONTOURS:strcpy(s, "Contours");  break;
	case T_GRAY:	strcpy(s, "Relief");    break;
	default:
	  strncpy(s, layerName[colorLayer],CONTROLLABELLENGTH);
	  break;
	}
	break;

      case 'e':  strcpy(s, eyeUp ? "Up" : "Down");	break;

      case 'f':
	strcpy(s, cullFOV ? "On" : "Off");  break;

      case '2':
	switch(twoDMesh) {
	case NONE2D:  strcpy(s, "Off");		break;
	case WIRE2D:  strcpy(s, "Outlined");	break;
	case FILL2D:  strcpy(s, "Filled");	break;
	}
	break;

      case '3':  strcpy(s, threeDMesh ? "On" : "Off");  break;

      case 'v':
	sprintf(s, "%3.1f", vexagger);  break;
	break;

      }
    }
  }

  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (controls[i].variable) {
      glColor3d((GLdouble)1,(GLdouble)0,(GLdouble)0);
    } else {
      glColor3d((GLdouble)0,(GLdouble)0,(GLdouble)0);
    }

    emitString(controls[i].x, controls[i].y, controls[i].label);

#ifdef SHOWCONTROLRECTANGLE
    glBegin(GL_LINE_LOOP);
    glColor3d((GLdouble)0,(GLdouble)0,(GLdouble)0);
    glVertex2i((GLint)2,(GLint)(controls[i].y + DSC));
    glVertex2i((GLint)(CONTROL_WIDTH - 3),(GLint)(controls[i].y + DSC));
    glVertex2i((GLint)(CONTROL_WIDTH - 3),(GLint)(controls[i].y-LSP+DSC));
    glVertex2i((GLint)2,(GLint)(controls[i].y-LSP+DSC));
    glEnd();
#endif

  }
  glFlush();
}



/* emitString(x, y, s) -- display a string  */

static void emitString(int x, int y, char *s)
{
  int len, i;

  glRasterPos2i((GLint)x,(GLint)y);
  len = strlen(s);
  for(i = 0; i < len; ++i) {
    glutBitmapCharacter(FONT, s[i]);
  }
}



/* controlPanelMouseFunc(btn, state, x, y) -- handle click in control window */

static void controlPanelMouseFunc(int button0, int state0, int x0, int y0)
{
  int i;

  if (state0 != GLUT_UP)
    return;

  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (y0 < controls[i].y + DSC && y0 > controls[i].y + DSC - LSP) {
      if (button0 == GLUT_LEFT_BUTTON)
	keyfunc(controls[i].cl, x0, y0);
      else if (button0 == GLUT_RIGHT_BUTTON)
	keyfunc(controls[i].cr, x0, y0);
      break;
    }
  }
}



/*
 * doloop() -- for timing redrawing and surface building
 */
#ifdef TESTING
static void doloop()
{
  int x,y,width,height;
  double starttime,endtime;
	
  glutSetWindow(ovhWin);
  width = glutGet(GLUT_WINDOW_WIDTH);
  height = glutGet(GLUT_WINDOW_HEIGHT);
  fprintf(stderr,"Window dimensions = %d x %d\n",width,height);
  x = 0;
  y = height-1;
	
  doingLoop = 1;
	
  starttime = getTime();
	
  for ( ;x < width; ++x)
    mouseto(GLUT_LEFT_BUTTON,x,y);
	
  for ( ;y >= 0; --y)
    mouseto(GLUT_LEFT_BUTTON,x,y);
    
  for ( ;x >= 0; --x)
    mouseto(GLUT_LEFT_BUTTON,x,y);
    
  for ( ;y < height; ++y)
    mouseto(GLUT_LEFT_BUTTON,x,y);
	
  endtime = getTime();

  if (doRedraw) {
    fprintf(stderr,"Loop with redraw in %f secs\n",endtime-starttime);
  } else {
    fprintf(stderr,"Loop without redraw in %f secs\n",endtime-starttime);
  }
	
  doingLoop = 0;
}
#endif


/*  keyfunc(key, x, y) -- handle keypress in any window  */

void keyfunc(unsigned char key, int x, int y)
     /*ARGSUSED*/
{
  static double stashedET = 0.0;
  int i;
  int sign = 1;
  int refresh = -1;
  char s[200];

  switch(key) {

#ifdef TESTING
  case 'L':
    doRedraw = 0;
    doloop();
    break;
  case 'l':
    doRedraw = 1;
    doloop();
    break;
#endif      

  case '<':
    sign = -sign;
    /*NOBREAK*/
  case '>':
    errThresh  += sign * errThreshInc;
    viewChanged();
    break;

  case '-':
    sign = -sign;
    /*NOBREAK*/
  case '+':
    distThresh += sign * distThreshInc;
    viewChanged();
    break;

  case 'B':
  case 'b':
    cullBackface = !cullBackface;
    break;

  case 'F':
  case 'f':
    cullFOV = !cullFOV;
    break;

  case 'c':
    sign = -sign;
    /* NOBREAK */
  case 'C':
    colorLayer += sign;
    if (colorLayer < T_GRAY) {
      colorLayer = nLayers-1;
    } else if (colorLayer >= nLayers) {
      colorLayer = T_GRAY;
    }
    break;

  case '2':
    switch(twoDMesh) {
    case NONE2D:  twoDMesh = WIRE2D;	break;
    case WIRE2D:  twoDMesh = FILL2D;	break;
    case FILL2D:  twoDMesh = NONE2D;	break;
    }
    break;

  case '@':
    switch(twoDMesh) {
    case WIRE2D:  twoDMesh = NONE2D;	break;
    case FILL2D:  twoDMesh = WIRE2D;	break;
    case NONE2D:  twoDMesh = FILL2D;	break;
    }
    break;

  case '0':
    if (stashedET == 0.0) {
      stashedET = errThresh;
      errThresh = 0.0;
    } else {
      errThresh = stashedET;
      stashedET = 0.0;
    }
    viewChanged();
    break;

  case 'E':
  case 'e':  setEye(!eyeUp);			break;

  case '#':
  case '3':  threeDMesh = !threeDMesh;	break;

  case 'V':
    sign = -sign;
    /* NOBREAK */
  case 'v':
    vexagger += sign * vexaggerInc;
    viewChanged();
    break;

  case 'P':
#ifndef NDEBUG
    outputEPS(1,NULL);
    break;
#endif
  case 'p':
    sprintf(s,"%s.ps",gemname);
    outputEPS(1,s);
    break;

  case 'Q':
  case 'q':  exit(0);				break;

  case ' ':  mmove(+gemXmax / 5.0, 0.0);		break;
  case '\b': mmove(-gemXmax / 5.0, 0.0);		break;

  default:  return;	/* unrecognized; no further processing */
  }

  /* search control panel table for refresh flags */
  for(i = 0; i < NUMBER_OF_CONTROLS; ++i) {
    if (controls[i].cl == key || controls[i].cr == key) {
      refresh = controls[i].refresh;
      break;
    }
  }

  /* refresh any necessary displays according to the flags set */
  if (refresh & RC) {
    glutSetWindow(ctlWin);
    glutPostRedisplay();
  }
  if (refresh & RO) {
    glutSetWindow(ovhWin);
    glutPostRedisplay();
  }
  if (refresh & RP) {
    glutSetWindow(pspWin);
    glutPostRedisplay();
  }
}



/*  specfunc(key, x, y) -- handle function key or other special key  */

void specfunc(int key, int x, int y)
     /*ARGSUSED*/
{
  switch (key) {
  case GLUT_KEY_UP:		mmove(+gemXmax / 20.0, 0.0);	break;
  case GLUT_KEY_DOWN:		mmove(-gemXmax / 20.0, 0.0);	break;
  case GLUT_KEY_LEFT:		mmove(0.0, +15.0);	break;
  case GLUT_KEY_RIGHT:		mmove(0.0, -15.0);	break;
  case GLUT_KEY_PAGE_UP:	setEye(1);		break;
  case GLUT_KEY_PAGE_DOWN:	setEye(0);		break;
  default:			/* nothing */		break;
  }
}



/*  mmove(step, turn) -- step forward and swivel gaze */

/*  if the shift key is down, scale both argument values by 0.25  */
/*  (call only from within a callback: it calls glutGetModifiers) */

static void mmove(double step, double turn)
{
  double dx, dy, d, a;

  if (glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
    step *= 0.25;
    turn *= 0.25;
  }

  dx = refEmx - eyeEmx;
  dy = refEmy - eyeEmy;
  d = sqrt(dx * dx + dy * dy);	/* current distance */
  a = atan2(dy, dx);			/* current view angle */

  eyeEmx += step * cos(a);		/* new eye location */
  eyeEmy += step * sin(a);
  refEmx += step * cos(a);		/* also translate reference point */
  refEmy += step * sin(a);

  a += .017453293 * turn;		/* new view angle (in radians) */
  refEmx = eyeEmx + d * cos(a);		/* new reference point */
  refEmy = eyeEmy + d * sin(a);

  viewChanged();			/* note redisplay needed */
}



/*  setEye(up) -- set eye height for current viewpoint  */
/*  calls PostRedisplay if up/down flag has changed  */

static void setEye(int up)
{
  if (up != eyeUp) {
    eyeUp = up;		/* save new value if different */
    viewChanged();		/* note that view changed */
    glutSetWindow(ctlWin);	/* rewrite control window to show new value */
    glutPostRedisplay();
  }
}

