/*  infov.c -- field-of-view (FOV) calculations  */

#include "topo.h"



/* A triangle is in FOV iff one of its 3 vertices is in FOV
 * or if the line of sight intersects one of its 3 segments.
 */
static int inFOV(Coord2d p1, Coord2d p2, Coord2d p3)
{
  int ret = ALLINFOV;
  double x1, y1, d1, x2, y2, d2, x3, y3, d3;
  double e1, e2, e3;

  x1 = p1[0] - eyeEmx;
  y1 = p1[1] - eyeEmy;
  d1 = eye2refx * x1 + eye2refy * y1;
  if (d1 > 0.0 && d1 * d1 / (x1 * x1 + y1 * y1) > inAngleThresh) {
    ret |= PARTINFOV;
  } else {
    ret &= ~ALLINFOV;
  }

  x2 = p2[0] - eyeEmx;
  y2 = p2[1] - eyeEmy;
  d2 = eye2refx * x2 + eye2refy * y2;
  if (d2 > 0.0 && d2 * d2 / (x2 * x2 + y2 * y2) > inAngleThresh) {
    ret |= PARTINFOV;
  } else {
    ret &= ~ALLINFOV;
  }

  x3 = p3[0] - eyeEmx;
  y3 = p3[1] - eyeEmy;
  d3 = eye2refx * x3 + eye2refy * y3;
  if (d3 > 0.0 && d3 * d3 / (x3 * x3 + y3 * y3) > inAngleThresh) {
    ret |= PARTINFOV;
  } else {
    ret &= ~ALLINFOV;
  }

  if (ret & ALLINFOV) {
    return ALLINFOV;
  }
  if (ret & PARTINFOV) {
    return PARTINFOV;
  }

  e1 = x1 * eye2refy - y1 * eye2refx;
  e2 = x2 * eye2refy - y2 * eye2refx;
  e3 = x3 * eye2refy - y3 * eye2refx;

  if ((((e1 >= 0.0 && e2 <= 0.0) || (e1 <= 0.0 && e2 >= 0.0)) &&
      (d1 > 0.0 || d2 > 0.0)) ||
      (((e2 >= 0.0 && e3 <= 0.0) || (e2 <= 0.0 && e3 >= 0.0)) &&
      (d2 > 0.0 || d3 > 0.0)) ||
      (((e3 >= 0.0 && e1 <= 0.0) || (e3 <= 0.0 && e1 >= 0.0)) &&
      (d3 > 0.0 || d1 > 0.0))) {
    return PARTINFOV;
  }

  return 0;
}



/*  triInFOV(t) -- is triangle t in field of view?  */

int triInFOV(TriIndex *t)
{
  Coord2d p1, p2, p3;

  getVerts(t, p1, p2, p3);
  return inFOV(p1, p2, p3);
}



/*  inROD(p1, p2, p3)  */

int inROD(Coord2d p1, Coord2d p2, Coord2d p3)
{
  int r;
  int txmax = MAX(p1[0], MAX(p2[0], p3[0]));
  int txmin = MIN(p1[0], MIN(p2[0], p3[0]));
  int tymax = MAX(p1[1], MAX(p2[1], p3[1]));
  int tymin = MIN(p1[1], MIN(p2[1], p3[1]));

  if (txmax > gemXmax - 1) {
    if (txmin >= gemXmax - 1) {
      r = 0;
    } else {
      r = PARTINROD;
    }
  } else if (tymax > gemYmax - 1) {
    if (tymin >= gemYmax - 1) {
      r = 0;
    } else {
      r = PARTINROD;
    }
  } else {
    r = ALLINROD;
  }

  return r;
}
