/*  overhead.c -- creation, drawing, and maintenance of overhead view window  */

#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <time.h>
#include "topo.h"

static void emitRelief(void);
static void emitViewfan(void);
static unsigned char *reliefImage(int w, int h);
static void mousefunc(int button, int state, int x, int y);
static void motionfunc(int x, int y);



static int whichBtn;			/* which mouse button is down? */



#define CMUL (255 * COLORATION)		/* color multiplier */
#define RMUL (255 * (1 - COLORATION))	/* relief multiplier */



/*  openOvhWin() -- initialize relief map window  */

void openOvhWin(void)
{
  char wname[200];
  int h = HEIGHT_2D;
  int w = HEIGHT_2D * xmeters / ymeters;
	
  sprintf(wname, "%s", gemname);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
  glutInitWindowSize(w, h);
  ovhWin = glutCreateWindow(wname);
  glutKeyboardFunc(keyfunc);
  glutSpecialFunc(specfunc);
  glutMouseFunc(mousefunc);
  glutMotionFunc(motionfunc);
  glutDisplayFunc(refreshOvh);
  buildSurface();			/* ensure valid triangle count */
}



/*  refreshOvh() -- refresh overhead display  */

void refreshOvh(void)
{
  char wname[200];
	
  if (twoDMesh != FILL3D) {
    emitRelief();
    emitViewfan();
  }
	
  if (twoDMesh == NONE2D) {
    sprintf(wname, "%s", gemname);
    glutSetWindowTitle(wname);
  } else {
    buildSurface();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluOrtho2D((GLdouble)0,(GLdouble)(gemXmax-1),(GLdouble)0,(GLdouble)(gemYmax-1));
    glColor3d((GLdouble)0,(GLdouble)0,(GLdouble)0);
    emitSurface(hierarchy, twoDMesh);
    sprintf(wname, "%d triangles", numDisplayTri);
    glutSetWindowTitle(wname);
  }
	
  glutSwapBuffers();
}



/*  emitRelief() -- display relief map  */

static void emitRelief(void)
{
  int w, h;
  static int wprev, hprev, cprev;
  static unsigned char *a;
	
  /* get current window size */
  w = glutGet(GLUT_WINDOW_WIDTH);
  h = glutGet(GLUT_WINDOW_HEIGHT);
	
  /* reconstruct image only if size or mode has changed */
  if (w != wprev || h != hprev || colorLayer != cprev)  {
    free(a);
    a = reliefImage(w, h);
    wprev = w;
    hprev = h;
    cprev = colorLayer;
  }
	
  /* draw the image */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  glDrawPixels(w, h, GL_RGB, GL_UNSIGNED_BYTE, a);
}



/*  emitViewfan() -- draw lines indicating viewpoint and direction  */

static void emitViewfan(void)
{
  int w, h;
  double ex, ey, rx, ry, dx, dy, d, a, a1, a2;
	
  /* get current window size */
  w = glutGet(GLUT_WINDOW_WIDTH);
  h = glutGet(GLUT_WINDOW_HEIGHT);
	
  /* place eye and reference points in window coordinates */
  ex = eyeEmx / gemXmax * w;
  ey = eyeEmy / gemYmax * h;
  rx = refEmx / gemXmax * w;
  ry = refEmy / gemYmax * h;
	
  /* calculate distance and angles */
  dx = rx - ex;
  dy = ry - ey;
  d = sqrt(dx * dx + dy * dy);
  a = atan2(dy, dx);
  a1 = a + 0.35;
  a2 = a - 0.35;
	
  /* draw lines */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluOrtho2D((GLdouble)0,(GLdouble)w,(GLdouble)0,(GLdouble)h);
  glColor3d((GLdouble)0,(GLdouble)0,(GLdouble)0);
  glBegin(GL_LINE_STRIP);
  glVertex2d((GLdouble)(ex + d * cos(a1)),(GLdouble)(ey + d * sin(a1)));
  glVertex2d((GLdouble)ex,(GLdouble)ey);
  glVertex2d((GLdouble)(ex + d * cos(a2)),(GLdouble)(ey + d * sin(a2)));
  glEnd();
}



/*  reliefImage(w, h) -- construct relief image for w-by-h window */

static unsigned char *reliefImage(int w, int h)
{
  Rgb *rgb;
  int *buf, *p;
  unsigned char *q;
  int i, j, k, x, y0, y1, d0, d1, z;
  int dz, dzmin, dzmax;
  double dx, dy, m, lum, r, g, b;
  Color *c;
  UV s;
	
  /* allocate buffer: used first for ints, then overlaid with rgb bytes */
  buf = (int *)malloc(w * h * sizeof(int));
  if (!buf)  {
    fprintf(stderr, "Can't allocate space for relief image.\n");
    exit(1);
  }
	
  /* compute step sizes and initial scaling factor */
  dx = gem->width / (double) w;
  dy = gem->height / (double) (h + 1);
	
  /* compute elevation difference for each output pixel */
  dzmin = dzmax = 0;
  p = buf;
  for (i = 0; i < h; i++) {
    y1 = gem->width * (int)(dy * (i + 1));
    if (dy < 1.0)
      y0 = y1 - gem->width;
    else
      y0 = y1 - gem->width * (int)(dy + 0.5);
    if (y0 < 0)
      y0 = 0;
    for (j = 0; j < w; j++) {
      x = dx * j;
      d0 = gem->data[y0 + x];
      d1 = gem->data[y1 + x];
      dz = d0 - d1;
      *p++ = dz;

      /* find min/max, omitting edges which have 0 (ob) data value */
      if (d0 != 0 && d1 != 0)  {
	if (dzmin > dz)
	  dzmin = dz;
	if (dzmax < dz)
	  dzmax = dz;
      }
    }
  }
	
  /* calculate multiplier for scaling to [-0.5, +0.5] */
  if (dzmax > -dzmin)
    m = 0.5 / dzmax;
  else if (dzmin == 0)
    m = 1.0;
  else
    m = -0.5 / dzmin;
	
  /* stretch contrast beyond that */
  m *= CONTRAST;
	
  /* compute r, g, b triples, overlaying atop old int values */
  p = (int *) buf;
  q = (unsigned char *) buf;
	
  if (colorLayer >= 0) {
    rgb = clayer[colorLayer];
  } else {
    rgb = NULL;
  }
  for (i = 0; i < h; i++) {
    y1 = gem->width * (int)(dy * (i + 1));  /* y offset in input em */
    for (j = 0; j < w; j++) {
			
      lum = 0.5 + m * *p++;		/* luminance */
      if (lum > 1.0)
	lum = 1.0;			/* clamp to valid range */
      if (lum < 0.0)
	lum = 0.0;
			
      if (rgb != NULL) {
	s.u = j * dx;
	s.v = i * dy;
	r = RED(rgb, s.u, s.v) / 255.0;
	g = GREEN(rgb, s.u, s.v) / 255.0;
	b = BLUE(rgb, s.u, s.v) / 255.0;
				
      } else if (colorLayer == T_CONTOURS) {
	x = dx * j;			/* x offset in input em */
	z = gem->data[y1+x];		/* elevation */
	if (z == 0) {
	  r = 0.4; g = 0.0; b = 0.0;	/* use dark red for non-data */
	} else if (z <= high_water && z >= low_water) {
	  r = 0.0;			/* use dark blue at sea level */
	  g = 0.0;
	  b = 0.4;
	} else {
	  k = z * dz2contour;		/* color index */
	  c = &elevMap[k];		/* color entry */
	  r = c->red;
	  g = c->green;
	  b = c->blue;
	}
				
      } else {
	r = g = b = lum;		/* use luminance only */
      }
			
      *q++ = RMUL * lum + CMUL * r;	/* r */
      *q++ = RMUL * lum + CMUL * g;	/* g */
      *q++ = RMUL * lum + CMUL * b;	/* b */
    }
  }
	
  /* discard unused portion of buffer, and return rgb bytes */
  return realloc(buf, 3 * w * h);
}



/*  mousefunc(btn, state, x, y) -- handle mouse click in overhead win */

static void mousefunc(int button, int state, int x, int y)
{
  whichBtn = button;			/* remember button for motion events */
  if (state != GLUT_UP)
    mouseto(button, x, y);		/* set position */
}



/*  motionfunc(x, y) -- handle mouse dragging  */

static void motionfunc(int x, int y)
{
  mouseto(whichBtn, x, y);
}



/*  mouseto(btn, x, y) -- handle mouse click or drag  */

#define DMS(x)	((int)(x))/3600, \
		ABS((int)((x)-3600*((int)(x)/3600))/60), \
		ABS((x)-60*((int)(x)/60))


void mouseto(int btn, int x, int y)
{
  UV mouseXY, mouseLL;
  int t,emxi,emyi,z;
  Rgb *rgb;
	
  mouseXY.u = gemXmax * (double)x / glutGet(GLUT_WINDOW_WIDTH);
  mouseXY.v = gemYmax * (1.0 - (double)y / glutGet(GLUT_WINDOW_HEIGHT));
  /* Y is reported as distance from top */
	
  if (btn == GLUT_RIGHT_BUTTON) {
    refEmx = mouseXY.u;
    refEmy = mouseXY.v;
    viewChanged();			/* note redisplay needed */
		
  } else if (btn == GLUT_LEFT_BUTTON) {
    eyeEmx = mouseXY.u;
    eyeEmy = mouseXY.v;
    viewChanged();			/* note redisplay needed */
		
  } else {
    emxi = (int)mouseXY.u;
    emyi = (int)mouseXY.v;
    z = zval(emxi,emyi);
    transform(gem->xy2ll,mouseXY,&mouseLL);
    printf("%3d %2d'%6.3f %3d %2d'%6.3f ",DMS(mouseLL.u),DMS(mouseLL.v));
    if (z==0) {
      printf("no elevation");
    } else {
      printf("elev %8.3fm",(z+gem->dzoffset)*gem->dz2m);
      if (colorLayer >= 0) {
	rgb = clayer[colorLayer];
	t = findLegend(rgb->legend,RED(rgb,emxi,emyi),
		       GREEN(rgb,emxi,emyi),BLUE(rgb,emxi,emyi));
	if( t < rgb->legend.size )
	  printf(" %s",rgb->legend.entry[t].label);
      }
    }
    printf("\n");
  }
	
}
