/*  split.c -- triangle splitting operations  */

#include "topo.h"



/* Triangle t is split into lt and rt.  Neighbors 1 & 2 of lt and
 * rt are known to be the same size as lt and rt.  Neighbor 3 of
 * lt is neighbor 2 of t (n3lt == n2t).  If n2t is smaller than
 * t then n3lt is same size as lt, o.w. n3lt is bigger than lt (and
 * n1n2t == lt is smaller than n2t == n3lt).
 * Similar gibberish holds for rt.
 */

static void kidsOnSurf(Hier *hp, TriIndex *t)
/* Push the surface down to the children of t.  t must not be a leaf. */
{
  TriIndex lt, rt, n2t, n1t;

  assert(!ISLEAF(hp,t));
  assert(t->s);
  assert(ONSURFACE(hp,t));

  hp->info[t->s] &= ~SURFACE;		/* take t from surface */
  hp->numSurfTri -= 1;
  leftChild(t,&lt);			/* put t's children on surface */
  if (lt.s) {
    PUTONSURFACE(hp,&lt);
    hp->numSurfTri += 1;
  }
  rightChild(t,&rt);
  if (rt.s) {
    PUTONSURFACE(hp,&rt);
    hp->numSurfTri += 1;
  }

  /* Let t's neighbors know they should point to t's children */
  nbr2(hp,t,&n2t);
  if (n2t.s && lt.s) {/* n2t.s && !(lt.s) could happen, but then n2t is invalid */
    assert(ONSURFACE(hp,&n2t));
    if (SMALLN2(hp,t)) {		/* n2t is same size as lt */
      hp->info[n2t.s] &= ~BIGGER3;
    } else {				/* n2t is bigger than lt */
      hp->info[n2t.s] |= SMALLER1;
      hp->info[lt.s] |= BIGGER3;
    }
  }
      
  nbr1(hp,t,&n1t);
  if (n1t.s && rt.s) {/* n1t.s && !(rt.s) could happen, but then n1t is invalid */
    assert(ONSURFACE(hp,&n1t));
    if (SMALLN1(hp,t)) {		/* n1t is same size as rt */
      hp->info[n1t.s] &= ~BIGGER3;
    } else {				/* n1t is bigger than lt */
      hp->info[n1t.s] |= SMALLER2;
      hp->info[rt.s] |= BIGGER3;
    }
  }

  if (hp->info[t->s] & ALLINFOV) {
    if (lt.s) hp->info[lt.s] |= ALLINFOV;
    if (rt.s) hp->info[rt.s] |= ALLINFOV;
  } else if (hp->info[t->s] & PARTINFOV) {
    if (lt.s) hp->info[lt.s] |= triInFOV(&lt);
    if (rt.s) hp->info[rt.s] |= triInFOV(&rt);
  }
}


void split(Hier *hp, TriIndex *t)
{
  TriIndex n3t,tmp;

  assert(t->s);
  assert(ONSURFACE(hp,t));

  nbr3(hp,t,&n3t);
  if (n3t.s == 0) goto N3INVALID;
  if (BIGN3(hp,t)) {

    split(hp, &n3t);	/* propagate split: t's 3-neighbor might change */

    if (IAMLEFTCHILD(t)) {
      rightChild(&n3t,&tmp);
    } else {
      leftChild(&n3t,&tmp);
    }
    if (tmp.s == 0) goto N3INVALID;
    TRIEQUATE(&n3t,&tmp);
  }
  kidsOnSurf(hp, &n3t);		/* t is not LEAF so n3t is not LEAF */

N3INVALID:
  kidsOnSurf(hp,t);
}
