/*  surface.c -- compute surface based on error criteria  */


#include "topo.h"

#define LEFT_TURN(a, b, c) \
(((b)[1]-(a)[1])*((c)[0]-(a)[0])-((b)[0]-(a)[0])*((c)[1]-(a)[1]) < 0)


static Coord2d eyeP;

static void rodSplit(Hier *hp, TriIndex *t)
{
  TriIndex lt,rt;

  if (t->s && !ISLEAF(hp, t) && BORDER(hp,t)) {
    if (ONSURFACE(hp,t)) {
      split(hp,t);
    }
    leftChild(t,&lt);
    rightChild(t,&rt);
    rodSplit(hp, &lt);
    rodSplit(hp, &rt);
  }
}


static void eyeSplit(Hier *hp, TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  Coord2d mid;
  TriIndex lt,rt;

  if (t->s && !ISLEAF(hp, t)) {
    if (ONSURFACE(hp,t)) {
      split(hp, t);
    }
    midpoint(p1, p2, mid);
    leftChild(t,&lt);
    rightChild(t,&rt);
    if (LEFT_TURN(mid, p3, eyeP)) {
      eyeSplit(hp, &lt, p3, p1, mid);
      rodSplit(hp, &rt);
    } else {
      eyeSplit(hp, &rt, p2, p3, mid);
      rodSplit(hp, &lt);
    }
  }
}

/* baseSurface --
 * calculates surface with finest level of detail around eye position
 * and no triangles spanning border of rectangle of definition.
 */
static void baseSurface(Hier *hp)
{
  eyeP[0] = (int)eyeEmx;
  eyeP[1] = (int)eyeEmy;

  /* SURFACE starts as triangles 1 and 2 */
  PUTONSURFACE(hp, &rootL);
  PUTONSURFACE(hp, &rootR);
  hp->numSurfTri = 2;
  hp->info[1] |= triInFOV(&rootL);
  hp->info[2] |= triInFOV(&rootR);


  if (LEFT_TURN(sw, ne, eyeP)) {
    eyeSplit(hp, &rootL, sw, ne, nw);	/* NW corner of square */
    rodSplit(hp, &rootR);
  } else {
    eyeSplit(hp, &rootR, ne, sw, se);	/* SE corner of square */
    rodSplit(hp, &rootL);
  }
}


/*-------------------ERROR SURFACE-----------------------------*/

static double etGrid,dtGrid;

static int tooCoarse(Hier *hp, TriIndex *t,
		     Coord2d p1, Coord2d p2, Coord2d p3)
{
  int h,dx,dy,d;

  assert(t->s);

  if (!INFOV(hp,t))		/* if not in field of view */
    return 0;


  /* Check distance from eye */
  dx = (int)(p3[0] - eyeP[0]);		/* dx and dy in grid units */
  dy = (int)(p3[1] - eyeP[1]);
  d = dx * dx + dy * dy;
  if (d == 0)				/* if right below eye */
    return 1;

  d = (int)sqrt((double)d);		/* d in grid units */
  if (d  < dtGrid * (hp->depth - t->depth))
    return 1;


  /* Check error threshold */
  h = hp->err[t->s];

  if (h > d * etGrid)
    return 1;

  return 0;
}


static void errSplit(Hier *hp, TriIndex *t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  TriIndex lt,rt;
  Coord2d mid;

  if (t->s && !ISLEAF(hp, t) && tooCoarse(hp, t, p1, p2, p3)) {
    if (ONSURFACE(hp, t)) {
      split(hp, t);
    }
    midpoint(p1, p2, mid);
    leftChild(t,&lt);
    rightChild(t,&rt);
    errSplit(hp, &lt, p3, p1, mid);
    errSplit(hp, &rt, p2, p3, mid);
  }
}


void errSurface(Hier *hp)
{
  double starttime=0.0,endtime;


  if (tflag) {
    starttime=getTime();
  }

  baseSurface(hp);

  etGrid = errThresh / err2grid;	/* set error parameters */
  dtGrid = distThresh * m2grid;

  errSplit(hp, &rootL, sw, ne, nw);	/* NW corner of square */
  errSplit(hp, &rootR, ne, sw, se);	/* SE corner of square */

  if (tflag) {
    endtime=getTime();
    fprintf(stderr,"Surface built in %f seconds.\n",endtime-starttime);
  }
}
